#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pickle
from timeit import Timer
import unittest
import sys

# Insert the parent directory to sys path so schem is accessible even if it's not available system-wide
sys.path.insert(1, str(Path(__file__).parent.parent))

from schem.molecule import Molecule

num_subtests = 0
LAST_TEST_RESULTS = {}


def percent_diff_str(last_metric, cur_metric):
    '''Provide a % indicator, colored red or green if the given metric significantly changed.'''
    percent_diff = (cur_metric - last_metric) / last_metric
    percent_formatting = ''
    if percent_diff <= -0.05:
        percent_formatting = '\033[92m'  # green
    elif percent_diff >= 0.05:
        percent_formatting = '\033[91m'  # red

    if percent_diff >= 0:
        percent_formatting += '+'

    format_end = '\033[0m'

    return f'{percent_formatting}{percent_diff:.0%}{format_end}'


class TestMolecule(unittest.TestCase):
    def test_isomorphism_valid(self):
        '''Tests for molecules that should be valid isomorphisms of each other.'''
        test_mol_pairs = (
            ("Wall of carbon",
             """Carbon;C;00611;10611;20611;30601;01611;12611;11611;21611;31601;32601;22611;02611;03611;13611;23611;33601;34601;24611;14611;04611;05611;15611;25611;35601;36601;26611;16611;06611;07610;17610;27610;37600""",
             "Carbon;C;00611;10611;20611;30601;01611;12611;11611;21611;31601;32601;22611;02611;03611;13611;23611;33601;34601;24611;14611;04611;05611;15611;25611;35601;36601;26611;16611;06611;07610;17610;27610;37600"),
            ("Doping rotated",
             """Silicon;Si;001411;111511;101411;011411;201411;301401;211411;311401;021411;121411;221411;321401;031410;131410;231410;331400""",
             """Silicon;Si;001411;111411;101411;011411;201411;301401;211411;311401;021411;121411;221511;321401;031410;131410;231410;331400"""),
            ("Pyramid",
             "Pyramid;C~01~06;12611;13611;14610;35601;32601;03610;34601;25610;36600;33601;22611;23611;24611;31601;30601;21611",
             "Pyramid;C~01~06;33601;22611;23611;24611;31601;30601;21611;12611;13611;14610;35601;32601;03610;34601;25610;36600"),)
        global num_subtests

        for test_id, mol_A_str, mol_B_str in test_mol_pairs:
            test_id = 'test_isomorphism_valid: ' + test_id
            with self.subTest(msg=test_id):
                num_subtests += 1
                mol_A = Molecule.from_json_string(mol_A_str)
                mol_B = Molecule.from_json_string(mol_B_str)

                self.assertTrue(mol_A.isomorphic(mol_B), "Molecule.isomorphic() unexpectedly returned False")

                # Check the time performance of the isomorphism algorithm
                timer = Timer('mol_A.isomorphic(mol_B)', globals={'mol_A': mol_A, 'mol_B': mol_B})

                if test_id not in LAST_TEST_RESULTS:
                    loops = max(max(timer.autorange()[0] for _ in range(2)) // 2, 5)
                else:
                    loops = LAST_TEST_RESULTS[test_id]['loops']

                min_time = min(timer.repeat(repeat=loops, number=1))

                if test_id in LAST_TEST_RESULTS:
                    metrics_str = f"{min_time:.5f}s (b. of {loops}) ({percent_diff_str(LAST_TEST_RESULTS[test_id]['min_time'], min_time)})"
                    metrics_str += f" | {test_id}"
                else:
                    metrics_str = f"{min_time:.5f}s (b. of {loops}) (NEW) | {test_id}"
                print(metrics_str)

                LAST_TEST_RESULTS[test_id] = {'loops': loops, 'min_time': min_time}


if __name__ == '__main__':
    test_results_file = Path(__file__).parent / 'last_performance_results.pickle'
    if test_results_file.exists():
        with test_results_file.open('rb') as f:
            LAST_TEST_RESULTS = pickle.load(f)

    unittest.main(verbosity=0, failfast=True, exit=False)
    print(f"Ran {num_subtests} subtests")

    # If successful, write the current times/mem usage to file
    with test_results_file.open('wb') as f:
        pickle.dump(LAST_TEST_RESULTS, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base64
import gzip
import io
import json
import zlib

from .molecule import Molecule

OVERWORLD_ROWS = 22
OVERWORLD_COLS = 32


# TODO: To reduce the complexity of Solution.__init__, provide something like a Level.default_components() method
class Level:
    """Parent class for Research and Production levels. Level(code) will return an instance of whichever subclass
    is appropriate.
    """
    export_line_len = 74
    __slots__ = 'dict', 'resnet_id'

    @classmethod
    def code_to_json(cls, code):
        try:
            return json.loads(zlib.decompress(base64.b64decode(code), wbits=16+15).decode('utf-8'))
        except Exception as e:
            raise ValueError("String is not a valid-format SpaceChem level code") from e

    def __new__(cls, code=None):
        """Return an instance of ResearchLevel or ProductionLevel as appropriate based on the given level code."""
        # If this is being called from a child class, behave like a normal __new__ implementation (to avoid recursion)
        if cls != Level or code is None:
            return object.__new__(cls)

        d = cls.code_to_json(code)
        # TODO: There doesn't seem to be a clean way to have Level(code) return an instance of the appropriate subclass,
        #       while also allowing ResearchLevel(code) and ProductionLevel(code) to work, without duplicating some
        #       conversion to/from the level code or else duplicating a lot of init logic.
        if d['type'].startswith('research'):
            return super().__new__(ResearchLevel)
        elif d['type'].startswith('production'):
            return super().__new__(ProductionLevel)
        elif d['type'] == 'sandbox':
            return super().__new__(SandboxLevel)
        elif d['type'] == 'defense':
            return super().__new__(DefenseLevel)
        else:
            raise ValueError(f"Unrecognized level type `{d['type']}`")

    # TODO: Allow initializing directly from dict too
    def __init__(self, code=None):
        if code is None:
            self.dict = {}
            self['name'] = 'Unknown'
            self['author'] = 'Unknown'
            self['difficulty'] = 0
        else:
            self.dict = self.code_to_json(code)

        # Since unfortunately some official levels share names, sometimes we want to track which volume/issue/puzzle the
        # level is, if it's a ResNet level.
        self.resnet_id = None

    def __eq__(self, other):
        return isinstance(other, Level) and self.dict == other.dict

    def __getitem__(self, item):
        return self.dict[item]

    def __setitem__(self, item, val):
        self.dict[item] = val

    def __contains__(self, item):
        return item in self.dict

    def __str__(self):
        return json.dumps(self.dict)

    def __delitem__(self, item):
        return self.dict.__delitem__(item)

    @property
    def code(self):
        """Export to mission code string; zip then b64 the level json."""
        out = zlib.compress(json.dumps(self.dict).encode('utf-8'), level=9, wbits=16+15)
        code = base64.b64encode(out).decode()
        # Line-wrap the export code for readability
        return '\n'.join(code[i:i+self.export_line_len] for i in range(0, len(code), self.export_line_len))

    @property
    def name(self):
        return self.dict['name']
    @name.setter
    def name(self, s):
        assert isinstance(s, str), "Level name must be a string"
        self.dict['name'] = s

    @property
    def type(self):
        return self.dict['type']
    @type.setter
    def type(self, s):
        assert isinstance(s, str), "Level type must be a string"
        self.dict['type'] = s

    # TODO: More properties and update Solution to use them


# TODO: Don't really need the subclasses by level type
class ResearchLevel(Level):
    __slots__ = ()

    def __init__(self, code=None):
        super().__init__(code)

        if code is None:
            self['type'] = 'research'
            self['input-zones'] = {}
            self['output-zones'] = {}

            self['has-large-output'] = False

            # Features of the level
            self['bonder-count'] = 0
            self['has-sensor'] = False
            self['has-fuser'] = False
            self['has-splitter'] = False
            self['has-teleporter'] = False

        assert self['type'].startswith('research')

    def output_molecules(self):
        """Return a list of Molecule objects demanded by this level."""
        return [Molecule.from_json_string(output_dict['molecule'])
                for _, output_dict in sorted(self['output-zones'].items())]


class ProductionLevel(Level):
    __slots__ = ()

    def __init__(self, code=None):
        super().__init__(code)

        if code is None:
            self['type'] = 'production'
            self['random-input-zones'] = {}
            self['fixed-input-zones'] = {}
            self['output-zones'] = {}
            self['terrain'] = 0
            self['max-reactors'] = 0
            self['has-starter'] = False
            self['has-assembly'] = False
            self['has-disassembly'] = False
            self['has-advanced'] = False
            self['has-nuclear'] = False
            self['has-recycler'] = False

        assert self['type'].startswith('production')


class SandboxLevel(Level):
    __slots__ = ()

    def __init__(self, code=None):
        super().__init__(code)

        if code is None:
            self['type'] = 'sandbox'
            self['random-input-zones'] = {}
            self['fixed-input-zones'] = {}
            self['programmed-input-start'] = []
            self['programmed-input-repeat'] = []
            self['programmed-input-molecules'] = {}

        assert self['type'] == 'sandbox'


class DefenseLevel(Level):
    __slots__ = ()

    def __init__(self, code=None):
        super().__init__(code)

        if code is None:
            self['type'] = 'defense'

            # Special defense level fields
            self['end-animation-cycles'] = 0  # The number of cycles that are added to the final score
            self['control-switches-allowed'] = True  # True for all defense levels except Collapsar
            self['max-cycles'] = None  # Player loses if this is exceeded; simulates player health

            self['random-input-zones'] = {}
            self['fixed-input-zones'] = {}
            self['weapons'] = {}
            self['max-reactors'] = 0
            self['has-starter'] = False
            self['has-assembly'] = False
            self['has-disassembly'] = False
            self['has-advanced'] = False
            self['has-nuclear'] = False
            self['has-recycler'] = False

        assert self['type'] == 'defense'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pickle
from timeit import Timer
import unittest
import sys

from pympler import asizeof

# Insert the parent directory to sys path so schem is accessible even if it's not available system-wide
sys.path.insert(1, str(Path(__file__).parent.parent))

import schem
import test_data

LAST_TEST_RESULTS = {}

num_subtests = 0


def iter_test_data(solution_codes):
    global num_subtests
    for solution_code in solution_codes:
        num_subtests += 1

        # Parse only the metadata line so we can error out from the appropriate subTest if the full parse fails
        level_name, _, _, solution_name = schem.Solution.parse_metadata(solution_code)
        test_id = f'{level_name} - {solution_name}' if solution_name is not None else level_name
        # Leave the level selection to Solution's constructor unless it's a custom level
        level_code = test_data.test_levels[level_name] if level_name in test_data.test_levels else None

        yield test_id, level_code, solution_code


def percent_diff_str(last_metric, cur_metric):
    '''Provide a % indicator, colored red or green if the given metric significantly changed.'''
    percent_diff = (cur_metric - last_metric) / last_metric
    percent_formatting = ''
    if percent_diff <= -0.05:
        percent_formatting = '\033[92m'  # green
    elif percent_diff >= 0.05:
        percent_formatting = '\033[91m'  # red

    if percent_diff >= 0:
        percent_formatting += '+'

    format_end = '\033[0m'

    return f'{percent_formatting}{percent_diff:.0%}{format_end}'


class TestPerformace(unittest.TestCase):
    def test_performance_valid_solutions(self):
        '''Tests for solutions that should run to completion and match the expected score.
        Also outputs runtime performance stats.
        '''
        for test_id, level_code, solution_code in iter_test_data(test_data.valid_solutions):
            with self.subTest(msg=test_id):
                # Sanity check that the test is passing
                solution = schem.Solution(solution_code, level=level_code)
                self.assertEqual(solution.run(), solution.expected_score)

                # TODO: This is only measuring the final object, we don't know the max runtime mem usage
                mem_usage = asizeof.asizeof(solution)

                # Check the time performance of the solver
                timer = Timer(('s=schem.Solution(solution_code, level=level_code)'
                               ';s.run()'),
                              globals={'level_code': level_code,
                                       'solution_code': solution_code,
                                       'schem': schem})

                if test_id not in LAST_TEST_RESULTS:
                    # autorange() measures how many loops of the solver needed to exceed 0.2 seconds. Ensures the slower
                    # solutions don't bloat the test times. We'll divide that down to 0.1s for expediency, or at least
                    # 5 runs per test, whichever is higher.
                    # Pretty meta, but since this number will stick for the lifetime of the test, I'm also measuring it
                    # twice and maxing it to be sure that a randomly extra-slow solve can't force us to under-sample a
                    # test forever. Running tests for the first time will take about an extra half second per new test.
                    loops = max(max(timer.autorange()[0] for _ in range(2)) // 2, 5)
                else:
                    # To avoid clogging our performance tests with tiny test cases which are all highly volatile anyway,
                    # if the last measured min_time was < 0.005s, skip it other than the above sanity check.
                    # Note that the original data will still be kept in the pickle (so we remember to skip it).
                    # Deleting the pickle file will prompt all tests to be re-measured once before being skipped for
                    # future runs.
                    if LAST_TEST_RESULTS[test_id]['min_time'] < 0.005:
                        continue

                    loops = LAST_TEST_RESULTS[test_id]['loops']

                # repeat = # of timeit calls, number = # of code executions per timeit call
                # I think they expect people to use autorange for `number` but the volatility seems to be better when
                # it's used for `repeat`. Probably repeat is better for large times, while number may help avoid the
                # measurement error of timeit outweighing short times.
                # Given our above loops measurement this should run in about 0.1s for lightweight tests
                min_time = min(timer.repeat(repeat=loops, number=1))

                if test_id in LAST_TEST_RESULTS:
                    metrics_str = f"{min_time:.4f}s (b. of {loops}) ({percent_diff_str(LAST_TEST_RESULTS[test_id]['min_time'], min_time)})"
                    metrics_str += f" | Mem usage: {mem_usage} B ({percent_diff_str(LAST_TEST_RESULTS[test_id]['mem_usage'], mem_usage)})"
                    metrics_str += f" | {test_id}"
                else:
                    metrics_str = f"{min_time:.4f}s (b. of {loops}) (NEW) | Mem usage: {mem_usage} B (NEW) | {test_id}"
                print(metrics_str)

                LAST_TEST_RESULTS[test_id] = {'loops': loops,
                                              'min_time': min_time,
                                              'mem_usage': mem_usage}


if __name__ == '__main__':
    # Use pickle to load the last test results and display a user-friendly summary of what tests performances went
    # up/down
    test_results_file = Path(__file__).parent / 'last_performance_results.pickle'
    if test_results_file.exists():
        with test_results_file.open('rb') as f:
            LAST_TEST_RESULTS = pickle.load(f)

    unittest.main(verbosity=0, failfast=True, exit=False)
    print(f"Ran {num_subtests} subtests")

    # If successful, write the current times/mem usage to file
    with test_results_file.open('wb') as f:
        pickle.dump(LAST_TEST_RESULTS, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import namedtuple


class Element(namedtuple("Element", ('atomic_num', 'symbol', 'max_bonds'))):
    """Immutable class representing a SpaceChemical element."""
    __slots__ = ()

    def __str__(self):
        return self.symbol
    __repr__ = __str__


elements = [Element(0, '?', 12),
            Element(1, 'H', 1),
            Element(2, 'He', 0),
            Element(3, 'Li', 1),
            Element(4, 'Be', 2),
            Element(5, 'B', 3),
            Element(6, 'C', 4),
            Element(7, 'N', 5),
            Element(8, 'O', 2),
            Element(9, 'F', 1),
            Element(10, 'Ne', 0),
            Element(11, 'Na', 1),
            Element(12, 'Mg', 2),
            Element(13, 'Al', 4),
            Element(14, 'Si', 4),
            Element(15, 'P', 5),
            Element(16, 'S', 6),
            Element(17, 'Cl', 7),
            Element(18, 'Ar', 0),
            Element(19, 'K', 1),
            Element(20, 'Ca', 2),
            Element(21, 'Sc', 3),
            Element(22, 'Ti', 4),
            Element(23, 'V', 5),
            Element(24, 'Cr', 6),
            Element(25, 'Mn', 7),
            Element(26, 'Fe', 6),
            Element(27, 'Co', 5),
            Element(28, 'Ni', 4),
            Element(29, 'Cu', 4),
            Element(30, 'Zn', 2),
            Element(31, 'Ga', 3),
            Element(32, 'Ge', 4),
            Element(33, 'As', 5),
            Element(34, 'Se', 6),
            Element(35, 'Br', 7),
            Element(36, 'Kr', 0),
            Element(37, 'Rb', 1),
            Element(38, 'Sr', 2),
            Element(39, 'Y', 3),
            Element(40, 'Zr', 4),
            Element(41, 'Nb', 5),
            Element(42, 'Mo', 6),
            Element(43, 'Tc', 7),
            Element(44, 'Ru', 8),
            Element(45, 'Rh', 6),
            Element(46, 'Pd', 4),
            Element(47, 'Ag', 3),
            Element(48, 'Cd', 2),
            Element(49, 'In', 3),
            Element(50, 'Sn', 4),
            Element(51, 'Sb', 5),
            Element(52, 'Te', 6),
            Element(53, 'I', 7),
            Element(54, 'Xe', 0),
            Element(55, 'Cs', 1),
            Element(56, 'Ba', 2),
            Element(57, 'La', 3),
            Element(58, 'Ce', 4),
            Element(59, 'Pr', 4),
            Element(60, 'Nd', 3),
            Element(61, 'Pm', 3),
            Element(62, 'Sm', 3),
            Element(63, 'Eu', 3),
            Element(64, 'Gd', 3),
            Element(65, 'Tb', 4),
            Element(66, 'Dy', 3),
            Element(67, 'Ho', 3),
            Element(68, 'Er', 3),
            Element(69, 'Tm', 3),
            Element(70, 'Yb', 3),
            Element(71, 'Lu', 3),
            Element(72, 'Hf', 4),
            Element(73, 'Ta', 5),
            Element(74, 'W', 6),
            Element(75, 'Re', 7),
            Element(76, 'Os', 8),
            Element(77, 'Ir', 6),
            Element(78, 'Pt', 6),
            Element(79, 'Au', 5),
            Element(80, 'Hg', 4),
            Element(81, 'Tl', 3),
            Element(82, 'Pb', 4),
            Element(83, 'Bi', 5),
            Element(84, 'Po', 6),
            Element(85, 'At', 7),
            Element(86, 'Rn', 0),
            Element(87, 'Fr', 1),
            Element(88, 'Ra', 2),
            Element(89, 'Ac', 3),
            Element(90, 'Th', 4),
            Element(91, 'Pa', 5),
            Element(92, 'U', 6),
            Element(93, 'Np', 7),
            Element(94, 'Pu', 7),
            Element(95, 'Am', 6),
            Element(96, 'Cm', 4),
            Element(97, 'Bk', 4),
            Element(98, 'Cf', 4),
            Element(99, 'Es', 3),
            Element(100, 'Fm', 3),
            Element(101, 'Md', 3),
            Element(102, 'No', 3),
            Element(103, 'Lr', 3),
            Element(104, 'Rf', 4),
            Element(105, 'Db', 5),
            Element(106, 'Sg', 6),
            Element(107, 'Bh', 7),
            Element(108, 'Hs', 8),
            Element(109, 'Mt', 6),

            # Greek Elements
            Element(200, 'Θ', 12),
            Element(201, 'Ω', 12),
            Element(202, 'Σ', 12),
            Element(203, 'Δ', 12),

            # Australium
            Element(204, 'Av', 5)]  # 🦘


class ElementDict(dict):
    """Class for looking up elements by atomic number, augmented internally to allow lookup by atomic symbol too."""
    __slots__ = 'symbol_dict',

    def __init__(self, elements):
        super().__init__()
        self.symbol_dict = {}

        for element in elements:
            self[element.atomic_num] = self.symbol_dict[element.symbol] = element

    def __contains__(self, key):
        return super().__contains__(key) or key in self.symbol_dict

    def __getitem__(self, key):
        if isinstance(key, int):
            return super().__getitem__(key)
        elif isinstance(key, str):
            return self.symbol_dict[key]
        else:
            raise TypeError(f"Elements must be looked up by atomic # or symbol; received {key}")


elements_dict = ElementDict(elements)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy
from pathlib import Path
import unittest
import sys

# Insert the parent directory to sys path so schem is accessible even if it's not available system-wide
sys.path.insert(1, str(Path(__file__).parent.parent))

import schem
from schem.solution import Solution
import test_data

num_subtests = 0


def iter_test_data(solution_codes):
    global num_subtests
    for solution_code in solution_codes:
        num_subtests += 1

        # Parse only the metadata line so we can error out from the appropriate subTest if the full parse fails
        level_name, _, _, solution_name = schem.Solution.parse_metadata(solution_code)
        test_id = f'{level_name} - {solution_name}' if solution_name is not None else level_name
        # Leave the level selection to Solution's constructor unless it's a custom level
        level_code = test_data.test_levels[level_name] if level_name in test_data.test_levels else None

        yield test_id, level_code, solution_code


class TestSolution(unittest.TestCase):
    def test_parse_metadata_valid(self):
        """Tests for valid solution metadata lines."""
        in_out_cases = [("SOLUTION:level_name,author,0-0-0", ("level_name", "author", None, None)),
                        ("SOLUTION:level_name,author,1-0-0", ("level_name", "author", (1, 0, 0), None)),
                        ("SOLUTION:level_name,author,0-0-0,soln_name", ("level_name", "author", None, "soln_name")),
                        ("SOLUTION:'commas,in,level,name',author,0-0-0,soln_name",
                         ("commas,in,level,name", "author", None, "soln_name")),
                        ("SOLUTION:trailing_field_quote',author,0-0-0,soln_name",
                         ("trailing_field_quote'", "author", None, "soln_name")),
                        ("SOLUTION:level_name,'commas,in,author,name',0-0-0,soln_name",
                         ("level_name", "commas,in,author,name", None, "soln_name")),
                        ("SOLUTION:level_name,author,0-0-0,unquoted,commas,in,soln name",
                         ("level_name", "author", None, "unquoted,commas,in,soln name")),
                        ("SOLUTION:quote_in_soln_name,author,0-0-0,''''",
                         ("quote_in_soln_name", "author", None, "'")),
                        ("SOLUTION:level_name,author,0-0-0,'comma , and quote '' in soln name'",
                         ("level_name", "author", None, "comma , and quote ' in soln name"))]

        for in_str, expected_outs in in_out_cases:
            with self.subTest(msg=in_str):
                self.assertEqual(tuple(Solution.parse_metadata(in_str)), expected_outs)

    def test_parse_metadata_invalid(self):
        """Tests for invalid solution metadata lines which should raise an exception."""
        invalid_cases = ("SOLUTION:level_name,'unescaped_leading_quote,0-0-0",)

        for invalid_metadata in invalid_cases:
            with self.subTest(msg=invalid_metadata):
                with self.assertRaises(Exception):
                    Solution.parse_metadata(invalid_metadata)

    def test_parse_metadata_legacy_formats(self):
        """Tests for solution metadata lines that were valid in legacy SC or community tools."""
        in_out_cases = [("SOLUTION:commas,in,level,name,author,0-0-0,soln_name",
                         ("commas,in,level,name", "author", None, "soln_name")),
                        ("SOLUTION:level_name,author,Incomplete-0-0,soln_name",
                         ("level_name", "author", None, "soln_name"))]

        for in_str, expected_outs in in_out_cases:
            with self.subTest(msg=in_str):
                self.assertEqual(tuple(Solution.parse_metadata(in_str)), expected_outs)

    def test_init_duplicate_level_name(self):
        """Test init on valid solutions to official levels with duplicate names."""
        for test_id, _, soln_str in iter_test_data(test_data.duplicate_level_name_solutions):
            with self.subTest(msg=test_id):
                schem.Solution(soln_str)
                print(f"✅  {test_id}")

    def test_init_errors(self):
        """Tests for solutions that shouldn't import successfully."""
        for test_id, level_code, solution_code in iter_test_data(test_data.import_errors):
            with self.subTest(msg=test_id):
                with self.assertRaises(schem.SolutionImportError):
                    schem.Solution(solution_code, level=level_code)
                print(f"✅  {test_id}")

    def test_init_duplicate_levels_import_error(self):
        """Test init with a solution which crashes on import to both same-name official levels."""
        with self.assertRaises(schem.SolutionImportError):
            soln_str = """SOLUTION:Sulfuric Acid,Zig,0-0-0,Unnamed Solution
COMPONENT:'custom-research-reactor',2,0,''"""
            schem.Solution(soln_str)

    def test_deepcopy(self):
        """Cursory test that deep copying a Solution works. Handy since we override __new__ a lot."""
        _, level_code, solution_code = next(iter_test_data(test_data.valid_solutions))
        copy.deepcopy(schem.Solution(solution_code, level=level_code))  # Just make sure this doesn't error

    def test_run_missing_score(self):
        """Test that run() does not require an expected score."""
        for test_id, level_code, solution_code in iter_test_data(test_data.missing_score):
            with self.subTest(msg=test_id):
                schem.Solution(solution_code, level=level_code).run()
                print(f"✅  {test_id}")

    def test_run_wrong_score(self):
        """Test that run() ignores whether the score does not match expected."""
        for test_id, level_code, solution_code in iter_test_data(test_data.wrong_score):
            with self.subTest(msg=test_id):
                schem.Solution(solution_code, level=level_code).run()
                print(f"✅  {test_id}")

    def test_run_runtime_collisions(self):
        """Tests for solutions that should encounter errors when run."""
        for test_id, level_code, solution_code in iter_test_data(test_data.runtime_collisions):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.SolutionRunError):
                    solution.run()
                print(f"✅  {test_id}")

    def test_run_wall_collisions(self):
        """Tests for solutions that should collide with a wall when run."""
        for test_id, level_code, solution_code in iter_test_data(test_data.wall_collisions):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.ReactionError) as context:
                    solution.run()
                self.assertTrue(' wall' in str(context.exception).lower())

                print(f"✅  {test_id}")

    def test_run_invalid_outputs(self):
        """Tests for solutions that should produce an InvalidOutput error."""
        for test_id, level_code, solution_code in iter_test_data(test_data.invalid_outputs):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.InvalidOutputError):
                    solution.run()
                    print(f"✅  {test_id}")

    def test_run_infinite_loops(self):
        """Tests for solutions that should have an infinite loop detected."""
        for test_id, level_code, solution_code in iter_test_data(test_data.infinite_loops):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.InfiniteLoopError):
                    solution.run()
                print(f"✅  {test_id}")

        # Ensure that this also work when a custom cycle handler is passed
        def handler(soln):
            soln.custom_data = 'whatever'

        for test_id, level_code, solution_code in iter_test_data(test_data.infinite_loops):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.InfiniteLoopError):
                    solution.run(cycle_handler=handler)
                print(f"✅  {test_id}")

    def test_run_pause(self):
        """Tests for solutions that should raise a PauseException and then succeed if the run is continued."""
        for test_id, level_code, solution_code in iter_test_data(test_data.pause_then_complete):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)

                # Run the solution and expect it to pause
                with self.assertRaises(schem.PauseException):
                    solution.run()

                # Make sure the displayed cycle on immediate pause matches SC's
                self.assertEqual(solution.cycle, 2, "Paused cycle does not match expected")

                # Make sure hitting run() again will complete the solution
                self.assertEqual(solution.run(), (154, 1, 11), "Solution failed to continue run after pause")

                print(f"✅  {test_id}")

    def test_run_encounter_ctrl(self):
        """Tests for solutions that should raise a ControlError for encountering a CTRL command when run."""
        for test_id, level_code, solution_code in iter_test_data(test_data.encounters_ctrl):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)

                # Run the solution and expect it to pause
                with self.assertRaises(schem.ControlError):
                    solution.run()

                print(f"✅  {test_id}")

    def test_validate_missing_score(self):
        """Test that validate() requires an expected score."""
        for test_id, level_code, solution_code in iter_test_data(test_data.missing_score):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(ValueError):
                    solution.validate()
                print(f"✅  {test_id}")

    def test_validate_wrong_score(self):
        """Test that validate() rejects successful solutions if the wrong score is specified."""
        for test_id, level_code, solution_code in iter_test_data(test_data.wrong_score):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.ScoreError):
                    solution.validate()
                print(f"✅  {test_id}")

    def test_validate_valid_solutions(self):
        """Tests for solutions that should run to completion and match the expected score."""
        for test_id, level_code, solution_code in iter_test_data(test_data.valid_solutions):
            with self.subTest(msg=test_id):
                schem.Solution(solution_code, level=level_code).validate()
                print(f"✅  {test_id}")

    def test_validate_custom_func(self):
        test_id, level_code, solution_code = next(iter_test_data(test_data.valid_solutions))
        solution = schem.Solution(solution_code, level=level_code)
        solution.custom_data = 0

        def custom_handler(soln):
            soln.custom_data += 1

        solution.validate(cycle_handler=custom_handler)
        self.assertEqual(solution.cycle, solution.custom_data, "Handler function not called")

    def test_death_solutions(self):
        """Tests for solutions which result in a loss on boss puzzles."""
        for test_id, level_code, solution_code in iter_test_data(test_data.death_solutions):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(schem.exceptions.DeathError):
                    solution.validate()
                self.assertEqual(solution.expected_score.cycles, solution.cycle, 'Death was not on the correct cycle')
                print(f"✅  {test_id}")

    def test_evaluate_duplicate_levels_success(self):
        """Test evaluate() with a solution which runs successfully in one of two same-name levels."""
        expected_json = {'level_name': "Sulfuric Acid",
                         'resnet_id': (3, 7, 1),
                         'cycles': 7208,
                         'reactors': 1,
                         'symbols': 36,
                         'author': "Zig",
                         'solution_name': "ResNet 3-7-1"}

        json = schem.Solution(test_data.duplicate_level_name_solutions[1]).evaluate()

        assert json == expected_json, f"Expected:\n{expected_json}\nbut got\n{json}"

    def test_evaluate_duplicate_levels_timeout_error(self):
        """Test evaluate() with a solution which imports successfully into one of two same-name
        levels, but times out.
        """
        expected_json = {'level_name': "Sulfuric Acid",
                         'resnet_id': (3, 7, 1),
                         'reactors': 1,
                         'symbols': 36,
                         'author': "Zig",
                         'solution_name': "ResNet 3-7-1"}

        json = schem.Solution(test_data.duplicate_level_name_solutions[1]).evaluate(max_cycles=10)

        assert 'cycles' not in json, "evaluate() unexpectedly set the cycles field after timeout"
        assert 'error' in json, "evaluate() failed to set error field"
        # Check all non-error fields have the expected values
        assert json.items() >= expected_json.items(), f"Expected:\n{expected_json}\nbut got\n{json}"

    def test_evaluate_duplicate_levels_cycles_score_error(self):
        """Test evaluate() with a solution which runs successfully in one of two same-name levels, but has a different
        cycle count than expected.
        """
        expected_json = {'level_name': "Sulfuric Acid",
                         'resnet_id': (3, 7, 1),
                         'cycles': 7208,  # Expect the actual cycle count to be returned
                         'reactors': 1,
                         'symbols': 36,
                         'author': "Zig",
                         'solution_name': "ResNet 3-7-1"}

        solution = schem.Solution(test_data.duplicate_level_name_solutions[1])
        solution.expected_score = schem.Score(7207, 1, 36)  # Understate cycle count by 1
        json = solution.evaluate()

        assert 'error' in json, "evaluate() failed to set error field"
        # Check all non-error fields have the expected values
        assert json.items() >= expected_json.items(), f"Expected:\n{expected_json}\nbut got\n{json}"

    def test_evaluate_duplicate_levels_symbols_score_error(self):
        """Test evaluate() with a solution which runs successfully in one of two same-name levels, but has a different
        symbol count than expected.
        """
        expected_json = {'level_name': "Sulfuric Acid",
                         'resnet_id': (3, 7, 1),
                         'reactors': 1,
                         'symbols': 36,  # Expect the actual symbol count to be returned
                         'author': "Zig",
                         'solution_name': "ResNet 3-7-1"}

        solution = schem.Solution(test_data.duplicate_level_name_solutions[1])
        solution.expected_score = schem.Score(7208, 1, 35)  # Understate symbol count by 1
        json = solution.evaluate()

        assert 'cycles' not in json, "evaluate() unexpectedly set the cycles field after symbols score failure"
        assert 'error' in json, "evaluate() failed to set error field"
        # Check all non-error fields have the expected values
        assert json.items() >= expected_json.items(), f"Expected:\n{expected_json}\nbut got\n{json}"

    def test_sandbox(self):
        """Test sandbox solutions load correctly and run to timeout (since they have no output components)."""
        for test_id, level_code, solution_code in iter_test_data(test_data.sandbox_solutions):
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                with self.assertRaises(TimeoutError):
                    solution.run(max_cycles=100_000)
                print(f"✅  {test_id}")

    def test_reset(self):
        """Test that solutions can be re-run after calling reset(), and still validate correctly."""
        for test_id, level_code, solution_code in iter_test_data(test_data.valid_solutions):
            test_id = 'Reset ' + test_id
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                solution.run()
                solution.reset()
                solution.validate()
                print(f"✅  {test_id}")

    def test_export_str(self):
        """Test that solutions can be re-exported to string, and that the new string still validates correctly."""
        for test_id, level_code, solution_code in iter_test_data(test_data.valid_solutions):
            test_id = 'Export ' + test_id
            with self.subTest(msg=test_id):
                solution = schem.Solution(solution_code, level=level_code)
                soln_from_export = schem.Solution(solution.export_str(), level=level_code)
                soln_from_export.validate()
                print(f"✅  {test_id}")

    def test_export_str_sorted(self):
        """Test that solution export strings are sorted consistently."""
        solution_code = next(s for s in test_data.valid_solutions
                             if Solution.parse_metadata(s) == ("An Introduction to Bonding", "Zig", (74, 1, 40), "Cycles"))
        export = schem.Solution(solution_code).export_str()
        # Start instructions at top and sorted by col then row, tie-breaking arrows above non-arrows
        expected_export = """SOLUTION:An Introduction to Bonding,Zig,74-1-40,Cycles
COMPONENT:'tutorial-research-reactor-2',2,0,''
MEMBER:'instr-start',-90,0,128,1,7,0,0
MEMBER:'instr-start',0,0,32,0,0,0,0
MEMBER:'feature-bonder',-1,0,1,1,1,0,0
MEMBER:'feature-bonder',-1,0,1,1,2,0,0
MEMBER:'feature-bonder',-1,0,1,1,4,0,0
MEMBER:'feature-bonder',-1,0,1,1,5,0,0
MEMBER:'instr-arrow',0,0,64,1,2,0,0
MEMBER:'instr-bond',-1,0,128,1,2,0,0
MEMBER:'instr-input',-1,0,128,1,3,0,0
MEMBER:'instr-input',-1,0,128,1,4,0,0
MEMBER:'instr-grab',-1,1,128,1,5,0,0
MEMBER:'instr-input',-1,1,128,1,6,0,0
MEMBER:'instr-input',-1,0,128,2,2,0,0
MEMBER:'instr-input',-1,0,128,4,2,0,0
MEMBER:'instr-arrow',0,0,64,4,3,0,0
MEMBER:'instr-input',-1,0,128,4,3,0,0
MEMBER:'instr-arrow',-90,0,64,4,4,0,0
MEMBER:'instr-grab',-1,1,128,4,4,0,0
MEMBER:'instr-bond',-1,0,128,5,2,0,0
MEMBER:'instr-rotate',-1,0,128,5,3,0,0
MEMBER:'instr-input',-1,1,128,5,4,0,0
MEMBER:'instr-arrow',90,0,64,6,2,0,0
MEMBER:'instr-arrow',90,0,64,6,3,0,0
MEMBER:'instr-grab',-1,2,128,6,3,0,0
MEMBER:'instr-arrow',180,0,64,6,4,0,0
MEMBER:'instr-arrow',90,0,16,0,2,0,0
MEMBER:'instr-arrow',0,0,16,0,3,0,0
MEMBER:'instr-arrow',90,0,16,1,0,0,0
MEMBER:'instr-input',-1,0,32,1,0,0,0
MEMBER:'instr-grab',-1,1,32,1,1,0,0
MEMBER:'instr-arrow',180,0,16,1,2,0,0
MEMBER:'instr-bond',-1,0,32,1,2,0,0
MEMBER:'instr-bond',-1,0,32,1,3,0,0
MEMBER:'instr-arrow',90,0,16,1,4,0,0
MEMBER:'instr-grab',-1,1,32,1,4,0,0
MEMBER:'instr-arrow',0,0,16,1,5,0,0
MEMBER:'instr-bond',-1,1,32,1,5,0,0
MEMBER:'instr-arrow',90,0,16,2,3,0,0
MEMBER:'instr-input',-1,1,32,2,3,0,0
MEMBER:'instr-arrow',180,0,16,2,4,0,0
MEMBER:'instr-bond',-1,0,32,2,4,0,0
MEMBER:'instr-output',-1,0,32,2,5,0,0
MEMBER:'instr-arrow',180,0,16,3,4,0,0
MEMBER:'instr-grab',-1,2,32,3,4,0,0
MEMBER:'instr-arrow',-90,0,16,3,5,0,0
MEMBER:'instr-rotate',-1,1,32,3,5,0,0
PIPE:0,4,1
PIPE:1,4,2"""

        assert export == expected_export, f"Solution export incorrect or not sorted correctly"
        print(f"✅  Export strings are sorted")

if __name__ == '__main__':
    unittest.main(verbosity=0, failfast=True, exit=False)
    print(f"Ran {num_subtests} subtests")
